package org.rosuda.jrs;

import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

import javax.script.AbstractScriptEngine;
import javax.script.Bindings;
import javax.script.Invocable;
import javax.script.ScriptContext;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineFactory;
import javax.script.ScriptException;
import javax.script.SimpleBindings;

/**
 * An abstract ScriptEngine that implements Invocable methods in terms of ScriptEngineEx.call.
 * @author Martyn Tebby
 */
public abstract class BaseScriptEngine
		extends AbstractScriptEngine
		implements ScriptEngineEx {

	private static final int WRITER_SIZE = 10000;

	private ScriptEngineFactory factory;

	@Override
	public Object eval(Reader reader, ScriptContext context) throws ScriptException {
		Writer writer = new StringWriter(WRITER_SIZE);
		int c = 0;
		try {
			while((c = reader.read()) != -1) {
				writer.write(c);
			}
		}
		catch(IOException ioe) {
			throw new ScriptException(ioe);
		}
		String script = writer.toString();
		return eval(script, context);
	}

	@Override
	public Bindings createBindings() {
		return new SimpleBindings();
	}

	@Override
	public ScriptEngineFactory getFactory() {
		return factory;
	}

	void setFactory(ScriptEngineFactory factory) {
		this.factory = factory;
	}

	@Override
	public Object invokeMethod(Object thiz, String name, Object... args)
	throws ScriptException, NoSuchMethodException {
		return call(null, thiz, name, args);
	}

	@Override
	public Object invokeFunction(String name, Object... args)
    throws ScriptException, NoSuchMethodException {
		return call(null, null, name, args);
	}

	@Override
	public <T> T getInterface(Class<T> cls) {
		return getInterface(null, cls);
	}

	@Override
	public <T> T getInterface(Object thiz, Class<T> cls) {
		Object proxy = Proxy.newProxyInstance(
				cls.getClassLoader(), new Class[] { cls }, new MyHandler(thiz));
		return cls.cast(proxy);
	}

	/** Close the engine. */
	@Override
	protected void finalize() {
		close();
	}

	/** A private InvocationHandler that uses call. */
	private class MyHandler implements InvocationHandler {
		private Object obj;
		private MyHandler(Object obj) {
			this.obj = obj;
		}
		public Object invoke(Object proxy, Method method, Object[] args)
		throws Exception {
			Class<?> clazz = method.getDeclaringClass();
			if(clazz == Object.class) {
				return method.invoke(this, args);
			}
			if(clazz == ScriptEngine.class || clazz == Invocable.class || clazz == ScriptEngineEx.class) {
				return method.invoke(BaseScriptEngine.this, args);
			}
			return call(method.getReturnType(), obj, method.getName(), args);
		}
	}

}
